@@ -3,7 +3,7 @@
  * Debian run-parts program
  * Copyright (C) 1996 Jeff Noxon <jeff@router.patch.net>,
  * Copyright (C) 1996-1999 Guy Maor <maor@debian.org>
- * Copyright (C) 2002-2012 Clint Adams <clint@debian.org>
+ * Copyright (C) 2002-2020 Clint Adams <clint@debian.org>
  *
  * This is free software; see the GNU General Public License version 2
  * or later for copying conditions.  There is NO warranty.
@@ -39,6 +39,7 @@
 int test_mode = 0;
 int list_mode = 0;
 int verbose_mode = 0;
+int debug_mode = 0;
 int report_mode = 0;
 int reverse_mode = 0;
 int exitstatus = 0;
@@ -62,7 +63,8 @@ static char* regex_get_error(int errcode, regex_t *compiled);
 static void  regex_compile_pattern(void);
 static void  regex_clean(void);
 
-void error(char *format, ...)
+__attribute__((format (printf, 1, 2)))
+void error(const char *format, ...)
 {
   va_list ap;
 
@@ -78,10 +80,10 @@ void error(char *format, ...)
 
 void version()
 {
-  fprintf(stderr, "Debian run-parts program, version " PACKAGE_VERSION
+  printf( "Debian run-parts program, version " PACKAGE_VERSION
 	  "\nCopyright (C) 1994 Ian Jackson, Copyright (C) 1996 Jeff Noxon.\n"
 	  "Copyright (C) 1996,1997,1998,1999 Guy Maor\n"
-	  "Copyright (C) 2002-2012 Clint Adams\n"
+	  "Copyright (C) 2002-2020 Clint Adams\n"
 	  "This is free software; see the GNU General Public License version 2\n"
 	  "or later for copying conditions.  There is NO warranty.\n");
   exit(0);
@@ -90,11 +92,12 @@ void version()
 
 void usage()
 {
-  fprintf(stderr, "Usage: run-parts [OPTION]... DIRECTORY\n"
+  printf("Usage: run-parts [OPTION]... DIRECTORY [DIRECTORY ...]\n"
 	  "      --test          print script names which would run, but don't run them.\n"
 	  "      --list          print names of all valid files (can not be used with\n"
 	  "                      --test)\n"
 	  "  -v, --verbose       print script names before running them.\n"
+	  "  -d, --debug         print script names while checking them.\n"
 	  "      --report        print script names if they produce output.\n"
 	  "      --reverse       reverse execution order of scripts.\n"
 	  "      --exit-on-error exit as soon as a script returns with a non-zero exit\n"
@@ -116,7 +119,7 @@ void usage()
  */
 void set_umask()
 {
-  int mask, result;
+  unsigned int mask, result;
 
   result = sscanf(optarg, "%o", &mask);
   if ((result != 1) || (mask > 07777) || (mask < 0)) {
@@ -151,19 +154,30 @@ int valid_name(const struct dirent *d)
 
     s = (char *)&(d->d_name);
 
-    if (regex_mode == RUNPARTS_ERE)
+    if (regex_mode == RUNPARTS_ERE) {
         retval = !regexec(&customre, s, 0, NULL, 0);
+	if (debug_mode)
+	    fprintf(stderr, "\"%s\": customre %s\n", s, retval ? "pass" : "fail");
 
-    else if (regex_mode == RUNPARTS_LSBSYSINIT) {
+    } else if (regex_mode == RUNPARTS_LSBSYSINIT) {
 
-        if (!regexec(&hierre, s, 0, NULL, 0))
+        if (!regexec(&hierre, s, 0, NULL, 0)) {
             retval = regexec(&excsre, s, 0, NULL, 0);
+	    if (debug_mode)
+		fprintf(stderr, "\"%s\": hierre pass, excsre %s\n", s, retval ? "pass" : "fail");
 
-	else
+	} else {
             retval = !regexec(&tradre, s, 0, NULL, 0);
+	    if (debug_mode)
+		fprintf(stderr, "\"%s\": tradre %s\n", s, retval ? "pass" : "fail");
 
-    } else
+	}
+    } else {
         retval = !regexec(&classicalre, s, 0, NULL, 0);
+	if (debug_mode)
+	    fprintf(stderr, "\"%s\": classicalre %s\n", s, retval ? "pass" : "fail");
+
+    }
 
     return retval;
 }
@@ -347,7 +361,7 @@ void run_part(char *progname)
 
   if (WIFEXITED(result) && WEXITSTATUS(result)) {
     error("%s exited with return code %d", progname, WEXITSTATUS(result));
-    exitstatus = 1;
+    exitstatus = WEXITSTATUS(result);
   }
   else if (WIFSIGNALED(result)) {
     error("%s exited because of uncaught signal %d", progname,
@@ -400,7 +414,7 @@ static int open_tmpfile_rw(void)
 }
 #else
 static char tmpfile_path[] = "/tmp/run-parts.stdin.XXXXXX";
-static int cleanup_tmpfile(void)
+static void cleanup_tmpfile(void)
 {
 	unlink(tmpfile_path);
 }
@@ -423,7 +437,6 @@ static int open_tmpfile_rw(void)
 static int copy_stdin(void)
 {
   int fd;
-  const char *tmpdir;
   char buffer[4096];
   ssize_t bytes;
 
@@ -455,31 +468,131 @@ static int copy_stdin(void)
   return fd;
 }
 
+static int qsort_strcoll(const void *p1, const void *p2) {
+  /* The actual arguments to this function are "pointers to
+     pointers to char", but strcoll(3) arguments are "pointers
+     to char", hence the following cast plus dereference. */
+  return strcoll(*(const char **)p1, *(const char **)p2);
+}
+
+static int remove_dupes(char **list, int len) {
+	if (len == 0) {
+		return 0;
+	}
+	int j = 1;
+	for (int i = 1; i < len; i++) {
+		// compare each element with the one before
+		if (strcmp(list[j-1], list[i]) == 0) {
+			// found duplicate, free memory
+			free(list[i]);
+			continue;
+		}
+		// not a duplicate
+		list[j] = list[i];
+		j++;
+	}
+	return j;
+}
+
 /* Find the parts to run & call run_part() */
-void run_parts(char *dirname)
+void run_parts(char **dirnames)
 {
-  struct dirent **namelist;
-  char *filename;
-  size_t filename_length, dirname_length;
-  int entries, i, result;
   struct stat st;
 
-  /* dirname + "/" */
-  dirname_length = strlen(dirname) + 1;
-  /* dirname + "/" + ".." + "\0" (This will save one realloc.) */
-  filename_length = dirname_length + 2 + 1;
-  if (!(filename = malloc(filename_length))) {
-    error("failed to allocate memory for path: %s", strerror(errno));
+  /* 1st step: gather a list of all files in the given directories */
+
+  char **basenames = NULL;
+  int num_basenames = 0;
+  for (int i = 0; dirnames[i] != NULL; i++) {
+    DIR *dirfd = opendir(dirnames[i]);
+    if (!dirfd) {
+      error("failed to opendir %s: %s", dirnames[i], strerror(errno));
+      exit(1);
+    }
+    struct dirent *dp;
+    while ((dp = readdir(dirfd)) != NULL) {
+      if (!strcmp(dp->d_name, ".") || !strcmp(dp->d_name, "..")) {
+        continue;
+      }
+      if (!valid_name(dp)) {
+        continue;
+      }
+      num_basenames++;
+      if (!(basenames = realloc(basenames, num_basenames * sizeof(char *)))) {
+        error("failed to reallocate memory for basenames: %s", strerror(errno));
+        exit(1);
+      }
+      basenames[num_basenames - 1] = strdup(dp->d_name);
+    }
+  }
+
+  if (num_basenames == 0) {
+    // nothing to do
+    return;
+  }
+
+  /* 2nd step: sort that list with alphasort */
+
+  qsort(basenames, num_basenames, sizeof(char *), qsort_strcoll);
+
+  /* 3rd step: make the sorted list only contain unique elements */
+
+  num_basenames = remove_dupes(basenames, num_basenames);
+  if (!(basenames = realloc(basenames, num_basenames * sizeof(char *)))) {
+    error("failed to reallocate memory for basenames: %s", strerror(errno));
     exit(1);
   }
-  strcpy(filename, dirname);
-  strcat(filename, "/");
+  if (debug_mode) {
+    fprintf(stderr, "list of unique basenames:\n");
+    for (int i = 0; i < num_basenames; i++) {
+      fprintf(stderr, " - %s\n", basenames[i]);
+    }
+  }
 
-  /* scandir() isn't POSIX, but it makes things easy. */
-  entries = scandir(dirname, &namelist, valid_name, alphasort);
-  if (entries < 0) {
-    error("failed to open directory %s: %s", dirname, strerror(errno));
-    exit(1);
+  /* 4th step: for each of the basenames, loop over the given directories
+   *           and store the full path of the first directory in which a
+   *           file with that basename was found */
+  char **full_paths = NULL;
+  int num_full_paths = 0;
+  char *tmp_full_path = NULL;
+  int max_full_path_length = 0;
+  for (int i = 0; i < num_basenames; i++) {
+    // Try to find the filename in each of the given directories, first match
+    // wins.
+    size_t basename_length = strlen(basenames[i]);
+    for (int j = 0; dirnames[j] != NULL; j++) {
+      size_t dirname_length = strlen(dirnames[j]);
+      // check if we need to increase the buffer for the temporary full path
+      if (max_full_path_length < basename_length + dirname_length + 2) {
+        max_full_path_length = basename_length + dirname_length + 2;
+        if (!(tmp_full_path = realloc(tmp_full_path, max_full_path_length))) {
+          error("failed to reallocate memory for path: %s", strerror(errno));
+          exit(1);
+        }
+      }
+      strcpy(tmp_full_path, dirnames[j]);
+      strcpy(tmp_full_path + dirname_length, "/");
+      strcpy(tmp_full_path + dirname_length + 1, basenames[i]);
+      if (debug_mode)
+        fprintf(stderr, "checking %s... ", tmp_full_path);
+      int result = stat(tmp_full_path, &st);
+      if (result < 0) {
+        if (debug_mode)
+          fprintf(stderr, "not found\n");
+        // not found in this directory, keep searching
+        continue;
+      }
+      // this path exists, so do not search further directories for matches
+      num_full_paths++;
+      if (!(full_paths = realloc(full_paths, num_full_paths * sizeof(char *)))) {
+        error("failed to reallocate memory for full paths: %s", strerror(errno));
+        exit(1);
+      }
+      full_paths[num_full_paths - 1] = strdup(tmp_full_path);
+      if (debug_mode)
+        fprintf(stderr, "found\n");
+      break;
+    }
   }
 
   if (stdin_mode) {
@@ -490,23 +603,13 @@ void run_parts(char *dirname)
     }
   }
 
-  i = reverse_mode ? 0 : entries;
-  for (i = reverse_mode ? (entries - 1) : 0;
-       reverse_mode ? (i >= 0) : (i < entries); reverse_mode ? i-- : i++) {
-    if (filename_length < dirname_length + strlen(namelist[i]->d_name) + 1) {
-      filename_length = dirname_length + strlen(namelist[i]->d_name) + 1;
-      if (!(filename = realloc(filename, filename_length))) {
-	error("failed to reallocate memory for path: %s", strerror(errno));
-	exit(1);
-      }
-    }
-    strcpy(filename + dirname_length, namelist[i]->d_name);
-
-    strcpy(filename, dirname);
-    strcat(filename, "/");
-    strcat(filename, namelist[i]->d_name);
-
-    result = stat(filename, &st);
+  /* 5th step: process the list of full paths */
+  for (int i = reverse_mode ? (num_full_paths - 1) : 0;
+       reverse_mode ? (i >= 0) : (i < num_full_paths); reverse_mode ? i-- : i++) {
+    char *filename = full_paths[i];
+    if (debug_mode)
+      fprintf(stderr, "processing: %s\n", filename);
+    int result = stat(filename, &st);
     if (result < 0) {
       error("failed to stat component %s: %s", filename, strerror(errno));
       if (exit_on_error_mode) {
@@ -526,7 +629,7 @@ void run_parts(char *dirname)
 	    printf("%s\n", filename);
 	}
 	else {
-	  if (verbose_mode)
+	  if (verbose_mode) {
 	    if (argcount) {
 	      char **a = args;
 
@@ -537,6 +640,7 @@ void run_parts(char *dirname)
 	    } else {
 	      fprintf(stderr, "run-parts: executing %s\n", filename);
 	    }
+	  }
 	  run_part(filename);
 	  if (exitstatus != 0 && exit_on_error_mode) return;
 	}
@@ -561,10 +665,22 @@ void run_parts(char *dirname)
       }
     }
 
-    free(namelist[i]);
   }
-  free(namelist);
-  free(filename);
+
+  if (full_paths != NULL) {
+    for (int i = 0; i < num_full_paths; i++) {
+      free(full_paths[i]);
+    }
+    free(full_paths);
+  }
+
+  if (basenames != NULL) {
+    for (int i = 0; i < num_basenames; i++) {
+      free(basenames[i]);
+    }
+    free(basenames);
+  }
+
 }
 
 /* Process options */
@@ -582,6 +698,7 @@ int main(int argc, char *argv[])
       {"test", 0, &test_mode, 1},
       {"list", 0, &list_mode, 1},
       {"verbose", 0, 0, 'v'},
+      {"debug", 0, 0, 'd'},
       {"report", 0, &report_mode, 1},
       {"reverse", 0, &reverse_mode, 1},
       {"umask", 1, 0, 'u'},
@@ -601,7 +718,7 @@ int main(int argc, char *argv[])
       break;
     switch (c) {
     case 0:
-      if(option_index==10) { /* hardcoding this will lead to trouble */
+      if(option_index==11) { /* hardcoding this will lead to trouble */
         custom_ere = strdup(optarg);
       }
       break;
@@ -617,6 +734,9 @@ int main(int argc, char *argv[])
     case 'v':
       verbose_mode = 1;
       break;
+    case 'd':
+      debug_mode = 1;
+      break;
     case 'V':
       version();
       break;
@@ -627,25 +747,27 @@ int main(int argc, char *argv[])
   }
 
   /* We require exactly one argument: the directory name */
-  if (optind != (argc - 1)) {
+  if (optind == argc) {
     error("missing operand");
     fprintf(stderr, "Try `run-parts --help' for more information.\n");
     exit(1);
-  } else if (list_mode && test_mode) {
+  }
+
+  if (list_mode && test_mode) {
     error("--list and --test can not be used together");
     fprintf(stderr, "Try `run-parts --help' for more information.\n");
     exit(1);
-  } else {
-    catch_signals();
-    regex_compile_pattern();
-    run_parts(argv[optind]);
-    regex_clean();
-
-    free(args);
-    free(custom_ere);
-
-    return exitstatus;
   }
+
+  catch_signals();
+  regex_compile_pattern();
+  run_parts(argv+optind);
+  regex_clean();
+
+  free(args);
+  free(custom_ere);
+
+  return exitstatus;
 }
 
 /*
@@ -677,7 +799,7 @@ regex_compile_pattern (void)
                     REG_EXTENDED | REG_NOSUB)) != 0)
             pt_regex = &excsre;
 
-        else if ( (err = regcomp(&tradre, "^[a-z0-9][a-z0-9-]*$", REG_NOSUB))
+        else if ( (err = regcomp(&tradre, "^[a-z0-9][a-z0-9_-]*$", REG_NOSUB))
                     != 0)
             pt_regex = &tradre;
 
@@ -686,7 +808,7 @@ regex_compile_pattern (void)
         pt_regex = &classicalre;
 
     if (err != 0) {
-        fprintf(stderr, "Unable to build regexp: %s", \
+        fprintf(stderr, "Unable to build regexp: %s\n", \
                             regex_get_error(err, pt_regex));
         exit(1);
     }
